/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * tPBIN.C
 *
 * PROGRAMMER : Gregg A. Tavares
 *    VERSION : 00.022
 *    CREATED : 02/23/93
 *   MODIFIED : 11/30/94
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 *	THIS IS AN UNPUBLISHED WORK CONTAINING ECHIDNA'S CONFIDENTIAL AND
 *	PROPRIETARY INFORMATION. DISCLOSURE, USE OR REPRODUCTION WITHOUT
 *	WRITTEN AUTHORIZATION OF ECHIDNA IS PROHIBITED.
 *
 * DESCRIPTION
 *
 *  Room Types
 *
 *  Tileset Types
 *		
 *
 * HISTORY
 *	0.020	02/05/94 (dcc) - Added -m-SORT to leave layers UNSORTED.
 *	0.021	08/25/94 (dcc) - Support reading new 'ROOM''PLAY' chunks.
 *	0.022	11/30/94 (dcc) - Add calls to ReleaseXTRA().
 *			11/30/94 (dcc) - Add support for '+p' switch.
 *			11/30/94 (dcc) - Compiles cleanly with 32-bit Borland C++ 4.02.
 *
 * SEE ALSO
 *		tPBIN.DOC -- Word for Windows file.
 *	
 * TODO	
 *
*/

#include <echidna/platform.h>
#include "switches.h"

#include <ctype.h>
#include <limits.h>
#include <echidna/etypes.h>
#include <echidna/version.h>
#include <echidna/listfunc.h>
#include <echidna/datafile.h>
#include <echidna/argparse.h>
#include <echidna/eio.h>
#include <echidna/eerrors.h>
#include <echidna/xtramem.h>
#include <echidna/btume.h>
#include <echidna/strings.h>

/**************************** C O N S T A N T S ***************************/

#define GRID_TILES			1

/*
** Tile Flags (this must NOT be the same any of the TUME_bits below!)
*/
#define TFLG_CONVERT		0x10	/* Convert this tile to chars		*/

/*
** tUME Map TileFlag Bits (these must match the tUME.INI definitions).
*/
#define	TUME_XFLIP			0x40
#define	TUME_YFLIP			0x20

/******************************** T Y P E S *******************************/

/*
** TilesetInfo
**
** Used to hold runtime information about each tileset.
*/
typedef struct {
	BOOL		 Done;			// Tileset has been remapped
	UBYTE		*Flags;			// 1. Which tiles to save, then 2. which flip bits
	UWORD		*RemapTable;	// Remap table for writing rooms
} TilesetInfo;

/****************************** G L O B A L S *****************************/

long			 MinMem			= LONG_MAX;	/* Used to track min free mem */
long			 MinMemEMS      = (-1L);
long             MinMemXMS      = (-1L);

BOOL			 DebugSwitch	= FALSE;	/* Print EXTRA info			*/
char			*MapName;					/* Filename of Map			*/
char			 TSName[EIO_MAXPATH];		/* Filename of char(font)	*/

BOOL			 PackFont		= FALSE;
BOOL			 FileSizes		= FALSE;
BOOL			 OnlyUsedTiles	= FALSE;
BOOL			 fSortLayers	= TRUE;
int			 FirstTile		= 0;

int			 LayerDest[256];

/******************************* M A C R O S ******************************/

#define GET_TOGGLE_SWITCH(var,arg)	\
	{ if ((arg)) (var) = (SWITCH_VALUE((arg)) == '+'); }

/******************************* T A B L E S ******************************/

/***************************** R O U T I N E S ****************************/


/*********************************************************************
 *
 * WritePalette
 *
 * PURPOSE
 *		Writes palette data to a file.  Write 256 colors.
 *
 * INPUT
 *		name	: Name of map.
 *		cr		: Pointer to color palette of room.
 *
 * RETURN VALUE
 *		NONE.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void WritePalette (char *name, ColorRegister *cr)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "WritePalette";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		FILE			*fp;
		int			 i;
		char			 rmfile[EIO_MAXPATH];

		EIO_fnmerge (rmfile, "", name, ".pal");

		printf ("Writing Palette '%s'\n", rmfile);

		fp = fopen (rmfile, "wb");
		if (!fp)
		{
			FailMess ("Couldn't create file '%s'\n", rmfile);
		}

		for (i = 0; i < 256; i++)
		{
			putc (cr->red,   fp);
			putc (cr->green, fp);
			putc (cr->blue,  fp);
			cr++;
		}

		fclose (fp);

	}
} /* WritePalette */

/* Local data structures for OpenTileStorage(), GetIndexInTileStoreFromImage(),
	AppendToTileStore(), and CloseTileStorage(). */

typedef struct
{
	NodeType	Node;
	UBYTE		pData[1];
} TIL;

static ListType *plstTiles;
static int cxTiles, cyTiles, cxyTiles;
static UWORD cTiles;

/*********************************************************************
 *
 * OpenTileStorage
 *
 * PURPOSE
 *		Initialize the data structure to store the tiles.
 *
 *		We're storing them so we can check for duplicates.
 *
 *		We're using a linked list, but you may want to optimize
 *		this code (big order of magnitude speed up) by using
 *		a binary tree or some better search structure.
 *
 * INPUT
 *		int cxTile			: width of tile in pixels
 *		int cyTile			: height of tile in pixels
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		FALSE if OOM.
 *
 * HISTORY
 *		11/30/94 (dcc) - Created.
 *
*/
static BOOL OpenTileStorage(int cxTile, int cyTile)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "OpenTileStorage";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	if ((plstTiles = CreateList(NULL)) == NULL)
		return FALSE;

	cxTiles = cxTile;
	cyTiles = cyTile;
	cxyTiles = cxTile * cyTile;
	cTiles = 0;

	return TRUE;

} /* OpenTileStorage */


/*********************************************************************
 *
 * GetIndexInTileStoreFromImage
 *
 * PURPOSE
 *		See if the image <ppix> occurs somewhere in the tile storage.
 *
 *		If so, return then index of the occurrence in the store
 *		in <uIndex>. Tiles are numbered starting from 0.
 *
 * INPUT
 *		
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		TRUE if found in tile storage.
 *
 *		UWORD *uIndex		: index of where tile occured in storage
 *		UBYTE *uFlags		: flip bits if flipped compared to image in store
 *
 * HISTORY
 *		11/30/94 (dcc) - Created.
 *
*/
static BOOL GetIndexInTileStoreFromImage(UBYTE *ppix, UWORD *uIndex, UBYTE *uFlags)
{
	TIL *ptil;
	UWORD c = 0;
	UBYTE *pFlipX, *pFlipY, *pFlipXY;

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "GetIndexInTileStoreFromImage";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	ptil = Head(plstTiles);
	if (IsEOList(ptil))
		return FALSE;

	{
		int ix, iy;
		UBYTE *desttr, *destbl, *destbr;
		UBYTE *ub = ppix;

		pFlipX  = malloc(cxyTiles);
		pFlipY  = malloc(cxyTiles);
		pFlipXY = malloc(cxyTiles);

		desttr = pFlipX + cxTiles - 1;
		destbl = pFlipY + cyTiles * (cxTiles - 1);
		destbr = pFlipXY + cxyTiles - 1;

		for (iy = 0; iy < cyTiles; iy++)
		{
			for (ix = 0; ix < cxTiles; ix++)
			{
				*desttr-- = *ub;
				*destbl++ = *ub;
				*destbr-- = *ub;
				ub++;
			}
			desttr += cxTiles * 2;
			destbl -= cxTiles * 2;
		}
	}

	while (!IsEOList(ptil))
	{
		if (!memcmp(ptil->pData, ppix, cxyTiles))
		{
			*uIndex = c;
			*uFlags = 0;
			return TRUE;
		}
		if (!memcmp(ptil->pData, pFlipX, cxyTiles))
		{
			*uIndex = c;
			*uFlags = TUME_XFLIP;
			return TRUE;
		}
		if (!memcmp(ptil->pData, pFlipY, cxyTiles))
		{
			*uIndex = c;
			*uFlags = TUME_YFLIP;
			return TRUE;
		}
		if (!memcmp(ptil->pData, pFlipXY, cxyTiles))
		{
			*uIndex = c;
			*uFlags = TUME_XFLIP | TUME_YFLIP;
			return TRUE;
		}
		c++;
		ptil = Next(ptil);
	}

	free(pFlipX);
	free(pFlipY);
	free(pFlipXY);

	return FALSE;

} /* GetIndexInTileStoreFromImage */


/*********************************************************************
 *
 * AppendToTileStore
 *
 * PURPOSE
 *		Add the tile image <ppix> to the tile storage.
 *
 * INPUT
 *		
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		FALSE if OOM.
 *
 *		UWORD *uIndex		: index of where tile was saved in the storage
 *
 * HISTORY
 *		11/30/94 (dcc) - Created.
 *
*/
static BOOL AppendToTileStore(UBYTE *ppix, UWORD *uIndex)
{
	TIL *ptil;

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "AppendToTileStore";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	if ((ptil = CreateNode(sizeof(TIL)+cxyTiles, NULL)) == NULL)
		return FALSE;

	memcpy (ptil->pData, ppix, cxyTiles);

	AddTail(plstTiles, ptil);

	*uIndex = cTiles++;

	return TRUE;

} /* AppendToTileStore */


/*********************************************************************
 *
 * CloseTileStorage
 *
 * PURPOSE
 *		Free memory used by tile storage.
 *
 * INPUT
 *		NONE.
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		NONE.
 *
 * HISTORY
 *		11/30/94 (dcc) - Created.
 *
*/
static void CloseTileStorage(void)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "CloseTileStorage";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	DeleteList(plstTiles);

} /* CloseTileStorage */


/*********************************************************************
 *
 * RemapLikeTiles
 *
 * PURPOSE
 *		If there are more than one tilesets of the same tile type,
 *		this routine makes sure that all tiles get unique numbers.
 *
 * INPUT
 *		TilesetType *origts	: use this to name tiles of this UserType.
 *		BOOL fMerge				: if TRUE, merge duplicate tiles.
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *		08/04/93 (dcc) - Start with Flags[1] and RemapTable[1] so
 *						 that last tile in tileset will be referenced.
 *
 * SEE ALSO
 *
*/
void RemapLikeTiles (MapType *map, TilesetType *origts, BOOL fMerge)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "RemapLikeTiles";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		TilesetType	**tsptr;
		TilesetType	*ts;
		TilesetInfo	*tsi;
		UWORD		 remapcount = (UWORD) FirstTile;
		char		 tsname[EIO_MAXPATH];
		char		 ext[10];
		int			 fh;

		sprintf (ext, ".t%02x", origts->UserType);
		EIO_fnmerge (tsname, NULL, EIO_Name(origts->FileSpec), ext);

		printf ("Writing %dx%d Tileset '%s'\n", origts->Width, origts->Height, tsname);

		fh = EIO_WriteOpen (tsname);
		if (fh == (-1))
		{
			FailMess ("Couldn't create file '%s'\n", tsname);
		}

		if (fMerge)
		{
			if (!OpenTileStorage(origts->Width, origts->Height))
				FailMess("Out of memory: Initializing tile storage.\n");
		}

		/*
		** Generate Remap tables
		*/
		tsptr = map->Tilesets + 1;
		while (*tsptr)
		{
			UBYTE			*flags;
			UWORD			*rtbl;

			ts    = *tsptr;
			tsi   = ts->UserData;
			flags = tsi->Flags + 1;
			rtbl  = tsi->RemapTable + 1;

			if (!tsi->Done &&
				ts->UserType == origts->UserType &&
				ts->Width    == origts->Width &&
				ts->Height   == origts->Height)
			{
				tsi->Done = TRUE;

				{
					int			 i;
					int			 tile;
					TileDataGroup	*tdg = NULL;
					UBYTE			*tiledata;
					UWORD			 tilesize;

					tilesize = (UWORD) (ts->TMG.Width * ts->TMG.Height);

					if (ts->TMG.TileCount)
					{
						tdg      = Head (ts->TMG.TileGroupList);
						tiledata = ActivateXTRA (tdg->TileData);
						tile     = tdg->NumTiles;
					}
					
					for (i = 0; i < ts->TileCount; i++)
					{
						if (*flags & TFLG_CONVERT)
						{
							if (fMerge)
							{
								UWORD uIndex;
								UBYTE uFlags;

								if (!GetIndexInTileStoreFromImage(tiledata, &uIndex, &uFlags))
								{
									if (ts->GridedTilesRoom)
									{
										WarnMess ("CompositeTiles not Supported Yet.\n");
									}
									else
									{
										EIO_Write (fh, tiledata, tilesize);
									}

									uFlags = 0;
									if (!AppendToTileStore(tiledata, &uIndex))
										FailMess("Out of memory: tile storage.\n");
								}
								*rtbl = (UWORD) (uIndex + FirstTile);
								*flags = uFlags;
							}
							else
							{
								*rtbl = remapcount;
								*flags = 0;
								remapcount++;

								if (ts->GridedTilesRoom)
								{
									WarnMess ("CompositeTiles not Supported Yet.\n");
								}
								else
								{
									EIO_Write (fh, tiledata, tilesize);
								}
							}
						}
						if (tdg)
						{
							tile--;
							if (tile)
							{
								tiledata += tilesize;
							}
							else
							{
								ReleaseXTRA(tdg->TileData);
								tdg = Next (tdg);
								if (!EndOfList (tdg))
								{
									tile = tdg->NumTiles;
									tiledata = ActivateXTRA (tdg->TileData);
								}
							}
						}
						rtbl++;
						flags++;
					}
				}
			}
			tsptr++;
		}

		EIO_Close (fh);

		if (fMerge)
		{
			CloseTileStorage();
		}
	}
} /* RemapLikeTiles */


/*********************************************************************
 *
 * ConvertAndWriteTiles
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
int ConvertAndWriteTiles (MapType *map)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ConvertAndWriteTiles";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		TilesetType	**tsptr;
		TilesetType	*ts;
		TilesetInfo	*tsi;

		/*
		** Allocate a remaptable and flags table for each tileset
		*/
		tsptr = map->Tilesets + 1;
		while (*tsptr) {
			ts  = *tsptr;

			tsi = calloc (1, sizeof (TilesetInfo));
			if (!tsi) {
				OutOfMemErr ("WriteAllTiles:TilesetInfo\n");
			}

			ts->UserData = tsi;
			tsi->RemapTable = calloc (ts->TileCount + 1, sizeof (UWORD));
			tsi->Flags      = calloc (ts->TileCount + 1, sizeof (UBYTE));
			if (!tsi->RemapTable || !tsi->Flags)
			{
				OutOfMemErr ("WriteAllTile:ImageChars\n");
			}

			//
			// if not 'OnlyUsedTiles' then mark all tile to be saved
			//
			if (!OnlyUsedTiles)
			{
				int	t;
				UBYTE	*tflg;

				tflg = tsi->Flags + 1;
				for (t = 0; t < ts->TileCount; t++)
				{
					*tflg = TFLG_CONVERT;
					tflg++;
				}
			}
			tsptr++;
		}

		/*
		** if (OnlyUsedTiles) then scan all the rooms and mark
		** all tiles found to be saved.
		**
		*/
		if (OnlyUsedTiles)
		{
			RoomType	**roomptr;
			RoomType	*room;

			printf ("Finding used tiles\n");

			roomptr = map->Rooms + 1;
			while (*roomptr)
			{
				int		 y;
				int		 x;
				int		 l;
				PlotType	*floorplot;

				room = *roomptr;

				for (l = 0; l < room->LayerCount; l++)
				{
					for (y = 0; y < room->Height; y++)
					{
						floorplot = ((PlotType *)ActivateXTRA (room->Layers[l]->Plot[y]));
						for (x = 0; x < room->Width; x++)
						{
							UBYTE		 tilesetID;
							UWORD		 tileID;
							TilesetType	*ts;
							TilesetInfo	*tsi;

							tilesetID = floorplot->TilesetID;
							tileID    = floorplot->TileID;

							if (tilesetID)
							{
								ts  = map->Tilesets[tilesetID];
								tsi = ts->UserData;
								tsi->Flags[tileID] |= TFLG_CONVERT;
							}
							floorplot++;
						}
						ReleaseXTRA(room->Layers[l]->Plot[y]);
					}
				}
				roomptr++;
			}
		}

		/*
		** Generate Remap tables
		*/
		tsptr = map->Tilesets + 1;
		while (*tsptr)
		{
			ts  = *tsptr;
			tsi = ts->UserData;

			if (!tsi->Done)
			{
				RemapLikeTiles (map, ts, PackFont);
			}
			tsptr++;
		}

		/*
		** Exit if there are any errors so far.
		*/
		return (!ErrorCount);
	}
} /* ConvertAndWriteTiles */


/************************************  ************************************/
/************************************  ************************************/
/************************************  ************************************/
/************************************  ************************************/
/************************************  ************************************/


/*********************************************************************
 *
 * WriteAllRooms
 *
 * PURPOSE
 *		Write all Level Rooms
 *		Write all Parallex/Picture Rooms
 *		Write all Table Rooms
 *
 *		While doing this, create BLOCK/FLOOR table for all Level Rooms
 *		in map.
 *
 * INPUT
 *		map		: pointer to a MapType.
 *
 * RETURN VALUE
 *		FALSE = failure.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
int WriteAllRooms (MapType *map)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "WriteAllRooms";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		RoomType	**roomptr;
		RoomType	*room;

		roomptr = map->Rooms + 1;
		while (*roomptr) {
			room = *roomptr;

			/*
			** Skip Composite Tileset Rooms
			*/
			if (room->CompositeTSet) {
			} else {

				if (!room->Name || !strlen(room->Name)) {
					FailMess ("Room size (%d, %d, %d) has no name!\n", room->Width, room->Height, room->LayerCount);
				}

				WritePalette (EIO_Name (room->Name), room->ColorInfo.Colors);

				{
					UWORD	**rlayers;
					UBYTE	**flayers;
					int	 layers;
					WORD	 width;
					WORD	 height;
					int	 l;
					int	 x;
					int	 y;
					int		*files;
					int		*pfiles;
					char	 rmname[EIO_MAXPATH];
					char	 ext[10];

					layers = room->LayerCount;
					width  = room->Width;
					height = room->Height;

					printf ("Writing room '%s' (%d,%d)\n", room->Name, width, height);

					rlayers = calloc (layers, sizeof (UWORD **));
					flayers = calloc (layers, sizeof (UBYTE **));
					files   = calloc (layers, sizeof (int));
					pfiles  = calloc (layers, sizeof (int));
					if (!rlayers || !flayers || !files || !pfiles)
					{
						OutOfMemErr ("rlayers\n");
					}

					for (l = 0; l < layers; l++)
					{
						rlayers[l] = malloc (width * sizeof (UWORD));
						if (!rlayers[l])
						{
							OutOfMemErr ("rlayers (2)\n");
						}

						flayers[l] = malloc (width * sizeof (UBYTE));
						if (!flayers[l])
						{
							OutOfMemErr ("flayers (3)\n");
						}

						sprintf (ext, ".r%02x", l);
						EIO_fnmerge (rmname, NULL, EIO_Name(room->Name), ext);

						files[l] = EIO_WriteOpen (rmname);
						if (files[l] == (-1))
						{
							FailMess ("Couldn't create file '%s'\n", rmname);
						}

						if (FileSizes)
						{
							EIO_Write (files[l], &width, sizeof (width));
							EIO_Write (files[l], &height, sizeof (height));
						}

						sprintf (ext, ".f%02x", l);
						EIO_fnmerge (rmname, NULL, EIO_Name(room->Name), ext);

						pfiles[l] = EIO_WriteOpen (rmname);
						if (pfiles[l] == (-1))
						{
							FailMess ("Couldn't create file '%s'\n", rmname);
						}
					}

					for (y = 0; y < height; y++)
					{
						for (l = 0; l < layers; l++)
						{
							memset (rlayers[l], 0, width * sizeof (UWORD));
							memset (flayers[l], 0, width * sizeof (UBYTE));
						}
						for (l = 0; l < layers; l++)
						{
							PlotType	*plot;

							plot = ActivateXTRA (room->Layers[l]->Plot[y]);
							for (x = 0; x < width; x++)
							{
								UBYTE	tsetID;
								UWORD	tileID;
								UBYTE	pflags;

								tsetID = plot->TilesetID;
								tileID = plot->TileID;
								pflags = plot->PlotFlags;

								if (tsetID)
								{
									TilesetType	*ts;
									TilesetInfo	*tsi;
									int		 utype;
									int		 dest;

									ts    = map->Tilesets[tsetID];
									tsi   = ts->UserData;
									utype = ts->UserType;

									if (fSortLayers)
										dest = LayerDest[utype];
									else
										dest = l;

									if (dest >= 0 && dest < layers)
									{
										if (rlayers[dest][x])
										{
											WarnMess ("More than one tile of same type at (%d,%d) in room %s\n", x, y, room->Name);
										}
										else
										{
											rlayers[dest][x] = tsi->RemapTable[tileID];
											flayers[dest][x] = pflags ^ tsi->Flags[tileID];
										}
									}
									else
									{
										WarnMess ("Tile at (%d,%d,%d) in room '%s' of type %d discarded\n", x, y, l, room->Name, utype);
									}
								}
								plot++;
							}
							ReleaseXTRA(room->Layers[l]->Plot[y]);
						}

						for (l = 0; l < layers; l++)
						{
							EIO_Write (files[l],  rlayers[l], width * sizeof (UWORD));
							EIO_Write (pfiles[l], flayers[l], width * sizeof (UBYTE));
						}
					}

					MinMemFree ();

					for (l = 0; l < layers; l++)
					{
						EIO_Close (files[l]);
						EIO_Close (pfiles[l]);
						free (rlayers[l]);
						free (flayers[l]);
					}

					free (rlayers);
					free (flayers);
					free (files);
					free (pfiles);
				}
			}
			roomptr++;
		}

		MinMemFree ();

		return (!ErrorCount);
	}
} /* WriteAllRooms */


/*********************************************************************
 *
 * GetNextVal
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
char *GetNextVal (char *s, int *val)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "GetNextVal";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		int	value = 0;

		while (isdigit (*s))
		{
			value = value * 10 + ((*s) - '0');
			s++;
		}
		*val = value;

		return s;
	}
} /* GetNextVal */


/******************************** TEMPLATE ********************************/

#define	ARG_D		(newargs[ 0])
#define	ARG_F		(newargs[ 1])
#define ARG_G		(newargs[ 2])
#define ARG_K		(newargs[ 3])
#define ARG_M		(newargs[ 4])
#define	ARG_P		(newargs[ 5])
#define ARG_HELP	(newargs[ 6])
#define	ARG_MAPNAME	(newargs[ 7])
#define	ARG_IGNORE	(newargs[ 8])

#define PROG	"tPBin"

char Usage[] = "Usage: " PROG " {-<switches>} <MAPNAME>\n";

ArgSpec Template[] = {
	{ CHRTOGGLE_ARG,	"d",	"\td - Debugging ON\n", },
	{ CHRKEYWORD_ARG,	"f",	"\tf<firsttile> Set first tile number (see docs)\n", },
	{ CHRTOGGLE_ARG,	"g",	"\tg - Write Room Width/Height into layer files\n", },
	{ CHRTOGGLE_ARG,	"k",	"\tk - Save only tiles used in rooms\n", },
	{ CHRKEYWORD_ARG,	"m",	"\tm<tileMap>\n"
								"\t      example: -m0=0,1=1,3=2,2=3\n"
								"\t         this says that tiles of usertype 0 are written to\n"
								"\t         layer 0, tiles of usertype 1 to layer 1, tile of type 3\n"
								"\t         to layer 2 and tiles of type 2 to layer 3\n\n"
								"\t      example: -m-SORT\n"
								"\t         this says not to sort the layers\n", },
	{ CHRTOGGLE_ARG,	"p",	"\tp - Pack Tiles\n", },
	{ CHRSWITCH_ARG,	"?",	"", },
	{ STANDARD_ARG|REQUIRED_ARG|MULTI_ARG, "MAPNAME", "", },
	{ 0, NULL, NULL, },
};

/*********************************************************************
 *
 * main
 *
 * PURPOSE
 *		Parse command line and make call to DoConversion.
 *
*/
int main (
	int		argc,
	char	**argv
)
{
	int		 result   = FALSE;
	char	**newargs;
	MapType	*map = NULL;

	MinMemFree ();
	printf (PROG " Version %s Copyright (c) Echidna (this work is unpublished)\n", Version);
	newargs = argparse (argc, argv, Template);
	if (!newargs || SWITCH_VALUE(ARG_HELP)) {
		printf ("%s\n", GlobalErrMsg);
		printarghelp (Usage, Template);
		return (EXIT_FAILURE);
	}

	GET_TOGGLE_SWITCH (DebugSwitch,    ARG_D);
	GET_TOGGLE_SWITCH (PackFont,       ARG_P);
	GET_TOGGLE_SWITCH (FileSizes,      ARG_G);
	GET_TOGGLE_SWITCH (OnlyUsedTiles,  ARG_K);

	if (ARG_F)	FirstTile = atoi (ARG_F);

	if (!ARG_M)
	{
		FailMess ("You must specify -m option and parameters\n");
	}

	if (strcmp(ARG_M, "-SORT") == 0)
	{
		fSortLayers = FALSE;
	}
	else
	{
		char	*s;
		int	 tiletype;
		int	 layer;

		fSortLayers = TRUE;
		s = ARG_M;

		for (layer = 0; layer < 256; layer++)
		{
			LayerDest[layer] = -1;
		}

		do
		{
			s = GetNextVal (s, &tiletype);

			if (*s != '=')
			{
				FailMess ("Bad -m specification\n");
			}
			s++;
			if (!isdigit (*s))
			{
				FailMess ("Bad -m specification\n");
			}

			s = GetNextVal (s, &layer);

			if (tiletype < 0 || tiletype > 255)
			{
				FailMess ("tiletype %d out of range.  Must be 0 - 255\n");
			}
			if (layer < 0 || layer > 255)
			{
				FailMess ("layer %d out of range.  Must be 0 - 255\n");
			}

			LayerDest[tiletype] = layer;

			if (*s == ',')
			{
				s++;
			}
		} while (*s);
	}

	//
	// load all requested maps
	//
	{
		char	**tfiles;

		tfiles           = MULTI_ARGLIST(ARG_MAPNAME);

		MapName = *tfiles;
		strupr (MapName);

		while (*tfiles) {
			map = AppendtUMEMapTMGC (map, *tfiles);
			if (!map) {
				break;
			}
			tfiles++;
		}
	}

	if (map) {
		result = ConvertAndWriteTiles (map);
		if (result) {
			result = WriteAllRooms (map);
		}
	}

	PrintMinMemFree();

	if (ErrorCount) {
		fprintf (stdout, "Errors: %d\n", ErrorCount);
		result = FALSE;
	}

	if (WarnCount) {
		fprintf (stdout, "Warnings: %d\n", WarnCount);
	}

	if (GlobalErr) {
		fprintf (stdout, "Error: %s\n", GlobalErrMsg);
		result = FALSE;
	}

	return (result ? EXIT_SUCCESS : EXIT_FAILURE);
}

